/* Main C code for Ppdf.  Written by Assassin Softwarez
						  Justin Moon
						  Stuart Wilcox
						  William Shales
						  Michael Norris
   This code is used for the Beta GUI for Ppdf.  Following code
   has callbacks for button functionality (bmp opening) and repositioning
   of group controls.                                                     */
/*                            AppBuilder Photon Code Lib */
/*                                         Version 1.13X */

/* Standard headers */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <malloc.h>
#include <assert.h>
#include <ctype.h>
#include <signal.h>
#include <process.h>
#include <sys/kernel.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <mem.h>
#include <photon/PhRender.h>

/* Toolkit headers */
#include <Ph.h>
#include <Pt.h>
#include <Ap.h>

#define PX_IMAGE_MODULES
#define PX_BMP_SUPPORT
#define PX_JPG_SUPPORT

#include <photon/PxImage.h>

/* Local headers */
#include "globals.h"
#include "abimport.h"
#include "proto.h"

int UseShmem=1;
void * memory_allocate (long nbytes, int type);
void memory_free (void * memory, int type);

#define ISFILE 1
#define ISPAGE 2
#define ISEXIT 3
#define TEXT 0
#define NONE 0
#define IMAGE 1
#define INLINE 2

#define BOLD 1
#define ITALICS 2

/* Application Options string */
const char ApOptions[] =
	AB_OPTIONS ""; /* Add your options in the "" */

char cur_file[180];  //Array that holds last opened file path
int fd; // Pointer to shared memory that ewoks will use to send us messages from endor
PhPoint_t gResizeData;//Thar be magic
char Pathd [160];
FontDetails list [200];
int NumOfFonts;

int First_Run_Call(void);

void
Exiting_With_A_Smile(void)
{

	kill (child,SIGKILL);
	shm_unlink("Endor");
	shm_unlink("ShMem");
}

void
SetPageNum (int CurrentPage)
{
	PtArg_t args[5];
	char page_num[5];
	itoa(CurrentPage,&page_num,10);
	
	PtSetArg(&args[0],Pt_ARG_TEXT_STRING,&page_num,0);
	PtSetResources(ABW_Ppdf_txt_CurPg, 1, args);
}

/*  RESIZE LABEL FUNCTION
    Resizes label to size of area scroll.  This resizes the label to fit
    size of image.
*/
void 
ResizeLabel(void)
{
    PtArg_t arg[1];  //Array of Arguments
	PhArea_t Area;  //Area structure for the label
       
	Area.pos.x = (0);  //Setting top left corner to 0,0
    Area.pos.y = (0);

	/*gResizeData is global pointer that is set everytime the area
	  scroll is resized*/
    Area.size.w =gResizeData.x;  //Setting height / width to 
    Area.size.h =gResizeData.y; //size of area scroll
		   
    PtSetArg (&arg[0], Pt_ARG_AREA, &Area, 1);  //Set label's area
    PtSetResources (ABW_Ppdf_lbl_PDFText, 1, arg);
}

/* used in Ppdf yet...*/
int
InitFcn( int argc, char *argv[] )

	{
	int i,num=200;
	
	for (i=0;i<(strlen(argv[0])-5);i++)
	{
		Pathd[i]=argv[0][i];
	}
	Pathd[i]='\x0';
	/* eliminate 'unreferenced' warnings */
	argc = argc, argv = argv;

	FileOpen=0;
	
	atexit(Exiting_With_A_Smile);
	NumOfFonts=PfQueryFonts('a',PHFONT_ALL_FONTS,&list,num);	
	return( Pt_CONTINUE );
	}

/* PPDF_WND_CB_RESIZE function
   In this function we repostion all group controls.
   (ie: buttons and page num information) allowing for correct window postioning
   on resize*/
int
Ppdf_Wnd_CB_Resize( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )

	{
    PtArg_t arg[1];   // Argument list
    PtContainerCallback_t   *data;     // will have old and new sizes
	PhArea_t Area;   // Area for groups
    data = cbinfo->cbdata;  //Assign data new size
   
    Area.pos.x = 273;//(data->new_size.lr.x-130); //Set top left and bottom right
    Area.pos.y = 40; //Of page num information group
    Area.size.w = 120;	//Set width/height to preset dimensions
    Area.size.h = 24;

    //Set page number information group's area
    PtSetArg (&arg[0], Pt_ARG_AREA, &Area, 1);
    PtSetResources (ABW_Pt_NumGroup, 1, arg);
	
	Area.pos.x = 12;  // Set x/y for button goup
    Area.pos.y = 40;
    Area.size.w = 134;  //Set width/height for button group
    Area.size.h = 29;
   	
	//Set area for button group
	PtSetArg (&arg[0], Pt_ARG_AREA, &Area, 1);
    PtSetResources (ABW_Ppdf_grp_buttons, 1, arg);

	Area.pos.x = 154;//((data->new_size.lr.x/2)-50);  // Set x/y for button goup
    Area.pos.y = 40;
    Area.size.w = 134;  //Set width/height for button group
    Area.size.h = 29;
   	
	//Set area for button group
	PtSetArg (&arg[0], Pt_ARG_AREA, &Area, 1);
    PtSetResources (ABW_Ppdf_grp_Scale, 1, arg);

	Area.pos.x = 500;//((data->new_size.lr.x/2)-50);  // Set x/y for button goup
    Area.pos.y = 20;
    Area.size.w = 150;  //Set width/height for button group
    Area.size.h = 29;
   	
	//Set area for button group
	PtSetArg (&arg[0], Pt_ARG_AREA, &Area, 1);
    PtSetResources (ABW_Ppdf_lbl_Ppdf, 1, arg);

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	return( Pt_CONTINUE );

	}


/*PPDF_BTN_LASTPG_CB_ACTIVATE function (Callback)
  Called when last page button clicked.  Opens an image (predefined)
  *.bmp  */
int
Ppdf_Btn_LastPg_CB_Activate( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )

	{
	
	if (FileOpen==1)
	{

		SendBuff.m_Type=ISPAGE;
		SendBuff.m_PageNum=lastpage;
		cur_page=lastpage;

		memset(gounta,0,sizeof(struct PagePass));

		Send (child,&SendBuff,&m_size,sizeof(struct Proto),sizeof(m_size));
		Send (child,&SendBuff,&cur_page,sizeof(struct Proto),sizeof(cur_page));

		//print number of pages on GUI
		PdfText=(struct PagePass *) malloc((sizeof(struct PagePass) + (m_size * sizeof (struct m_Page))));

		memcpy(PdfText,gounta,(sizeof(struct PagePass) + (m_size * sizeof (struct m_Page))));

		First_Run_Call();
		SetPageNum(cur_page);
	}

	// eliminate 'unreferenced' warnings 
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;
		
	return( Pt_CONTINUE );

	}


/*PPDF_BTN_NEXTPG_CB_ACTIVATE function (Callback)
  Called when next page button clicked.  Opens an image (predefined)
  *.bmp  */
int
Ppdf_Btn_NextPg_CB_Activate( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )

	{
	if (FileOpen==1)
	{
	
		SendBuff.m_Type=ISPAGE;
		cur_page++;
		if (cur_page<=lastpage)
		{
			SendBuff.m_PageNum=cur_page;

			Send (child,&SendBuff,&m_size,sizeof(struct Proto),sizeof(m_size));
			Send (child,&SendBuff,&cur_page,sizeof(struct Proto),sizeof(cur_page));

			//print number of pages on GUI
			PdfText=(struct PagePass *) malloc((sizeof(struct PagePass) + (m_size * sizeof (struct m_Page))));

			memcpy(PdfText,gounta,(sizeof(struct PagePass) + (m_size * sizeof (struct m_Page))));

			//print number of pages on GUI
			SetPageNum(cur_page);

			First_Run_Call();
		}
		else
			cur_page=lastpage;
	}

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	return( Pt_CONTINUE );

	}


/*PPDF_BTN_PREVPG_CB_ACTIVATE function (Callback)
  Called when Prev page button clicked.  Opens an image (predefined)
  *.bmp  */
int
Ppdf_Btn_PrevPg_CB_Activate( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )

	{
	
	if (FileOpen==1)
	{

		SendBuff.m_Type=ISPAGE;
		cur_page--;
		if (cur_page>=1)
		{
			SendBuff.m_PageNum=cur_page;
			Send (child,&SendBuff,&m_size,sizeof(struct Proto),sizeof(m_size));
			Send (child,&SendBuff,&cur_page,sizeof(struct Proto),sizeof(cur_page));

			//print number of pages on GUI
			PdfText=(struct PagePass *) malloc((sizeof(struct PagePass) + (m_size * sizeof (struct m_Page))));

			memcpy(PdfText,gounta,(sizeof(struct PagePass) + (m_size * sizeof (struct m_Page))));

			//print number of pages on GUI
			SetPageNum(cur_page);

			First_Run_Call();
		}
		else
			cur_page=1;
	}

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	return( Pt_CONTINUE );

	}


/*PPDF_BTN_FIRSTPG_CB_ACTIVATE function (Callback)
  Called when first page button clicked.  Opens an image (predefined)
  *.bmp  */
int
Ppdf_Btn_FirstPg_CB_Activate( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )

	{

	if (FileOpen==1)
	{
	
		SendBuff.m_Type=ISPAGE;
		cur_page=1;
		SendBuff.m_PageNum=cur_page;

		Send (child,&SendBuff,&m_size,sizeof(struct Proto),sizeof(m_size));
		Send (child,&SendBuff,&cur_page,sizeof(struct Proto),sizeof(cur_page));

		//print number of pages on GUI
		PdfText=(struct PagePass *) malloc((sizeof(struct PagePass) + (m_size * sizeof (struct m_Page))));

		memcpy(PdfText,gounta,(sizeof(struct PagePass) + (m_size * sizeof (struct m_Page))));

		//print number of pages on GUI
		SetPageNum(cur_page);

		First_Run_Call();
	}

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	return( Pt_CONTINUE );

	}


/*PPDF_SCROLLCONT_CB_RESIZE function
  Sets global ResizeData variable for scroll container resize
*/
int
Ppdf_ScrollCont_CB_Resize( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )

	{
	PtContainerCallback_t *data;

	data=cbinfo->cbdata;
	
    gResizeData.x = data->new_size.lr.x;
	gResizeData.y = data->new_size.lr.y;
	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	return( Pt_CONTINUE );

	}
int
First_Run_Call(void)

    {

	PhPoint_t trans = { 0,0 }, textloc,imagesize;
	PhDim_t dim = { 600,800 };
	PmMemoryContext_t *img;
	short bytes = 3;
    int loopcounter;
	PhImage_t Image2;
	char FilePath [180];
	char FileName2 [180];
	int i,size;
	char font [20];
	PtArg_t args[5];  //See above
	PhImage_t * Image;  //Pointer to Image structure
    PxMethods_t methods;  //Methods structure


      /*Set all variables to 0 in methods structure*/
	memset (&Image2, 0, sizeof(PhImage_t));

	Image2.type = Pg_IMAGE_DIRECT_888;
	Image2.size = dim;
	Image2.image = PgShmemCreate(dim.w * dim.h * bytes,"ShMem");

	img=PmMemCreateMC(&Image2, &dim, &trans);
	PmMemStart(img);
	if (FileOpen==0)
		PgSetFillColor(Pg_GREY);
	else	
		PgSetFillColor(Pg_WHITE);

	PgSetStrokeColor(Pg_BLACK);
	
	PgSetTextColor(Pg_BLACK);
	PgDrawIRect (0,0,dim.w,dim.h,Pg_DRAW_FILL_STROKE);
	if (FileOpen==1)
	{

		for (loopcounter=0;loopcounter<PdfText->NumOfItems;loopcounter++)
		{
			
			if (PdfText->PageStuff[loopcounter].m_type==IMAGE)
			{
				memset (&methods,0,sizeof(PxMethods_t));
				methods.flags=PX_LOAD|PX_DIRECT_COLOR|PX_USECOLORS;
				methods.px_alloc=memory_allocate;
				methods.px_free=memory_free;

				strcpy(&FileName2,&PdfText->PageStuff[loopcounter].m_Structs.m_Image.FileName[0]);
			
				textloc.x = PdfText->PageStuff[loopcounter].m_Structs.m_Image.x;
				textloc.y = PdfText->PageStuff[loopcounter].m_Structs.m_Image.y;
				imagesize.x=PdfText->PageStuff[loopcounter].m_Structs.m_Image.w;
				imagesize.y=PdfText->PageStuff[loopcounter].m_Structs.m_Image.h;
				Image=PxLoadImage(&FileName2,&methods);
				if (Image!= NULL)
					PgDrawImage(Image->image, Image->type, &textloc, &imagesize, Image->bpl,0);
				
				strcpy(&FilePath,"rm ");
				strcat(&FilePath,&FileName2);
				system(&FilePath);
			}
			else if (PdfText->PageStuff[loopcounter].m_type==TEXT)
			{
				size=PdfText->PageStuff[loopcounter].m_Structs.m_Text.size;
				for (i=0;i<NumOfFonts;i++)
				{
					if(strcmp(&list[i].desc,&PdfText->PageStuff[loopcounter].m_Structs.m_Text.FontName)==0)
					{
						strcpy(&font,&list[i].stem);
						if (list[i].losize>size)
							size=list[i].losize;
						break;
					}
				}
				if (i==NumOfFonts)
					strcpy(&font,"time");

				strcat(&font,(itoa(size,&FileName2,10)));
				if(PdfText->PageStuff[loopcounter].m_Structs.m_Text.bold==BOLD)
					strcat(&font,"b");	
				if(PdfText->PageStuff[loopcounter].m_Structs.m_Text.italic==ITALICS)
					strcat(&font,"i");
				PgSetFont(&font);

				textloc.x = PdfText->PageStuff[loopcounter].m_Structs.m_Text.x;
				textloc.y = PdfText->PageStuff[loopcounter].m_Structs.m_Text.y;
				strcpy(&FileName2,PdfText->PageStuff[loopcounter].m_Structs.m_Text.text);
				PgDrawText(&FileName2,strlen(&FileName2),&textloc,0);
			}
			else
				PgDrawIRect(PdfText->PageStuff[loopcounter].m_Structs.m_Inline.x,
				PdfText->PageStuff[loopcounter].m_Structs.m_Inline.y,PdfText->PageStuff[loopcounter].m_Structs.m_Inline.w,
				PdfText->PageStuff[loopcounter].m_Structs.m_Inline.h,Pg_DRAW_STROKE);

		}
	}
	

	PmMemFlush(img,&Image2);
		
	PmMemStop(img);
	PmMemReleaseMC (img);
    //Assign image to label
	PtSetArg(&args[0],Pt_ARG_LABEL_DATA,&Image2,sizeof(Image2));
	PtSetResources(ABW_Ppdf_lbl_PDFText, 1, args);

	ResizeLabel();	//Resizes Label to size of scroll area

	free(PdfText);

    return (69);

    }

int
Ppdf_txt_CurPg_Activate_CB( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )

	{
	PtArg_t args[5];
	char  *page_num;
	
	if (FileOpen==1)
	{
		PtSetArg(&args[0],Pt_ARG_TEXT_STRING,&page_num,0);
		PtGetResources(ABW_Ppdf_txt_CurPg, 1, args);
		cur_page=atoi(page_num);
		if (cur_page>=1 && cur_page<=lastpage)
		{
			SendBuff.m_Type=ISPAGE;
			SendBuff.m_PageNum=cur_page;

			Send (child,&SendBuff,&m_size,sizeof(struct Proto),sizeof(m_size));
			Send (child,&SendBuff,&cur_page,sizeof(struct Proto),sizeof(cur_page));

			//print number of pages on GUI
			PdfText=(struct PagePass *) malloc((sizeof(struct PagePass) + (m_size * sizeof (struct m_Page))));

			memcpy(PdfText,gounta,(sizeof(struct PagePass) + (m_size * sizeof (struct m_Page))));

			First_Run_Call();
		}
	}
		

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	return( Pt_CONTINUE );

	}

void * memory_allocate (long nbytes,int type)
{
	if (type==PX_IMAGE && UseShmem)
		return (PgShmemCreate (nbytes,NULL));
	else
		return(calloc(1,nbytes));
}
void memory_free (void * memory, int type)
{
	if (type==PX_IMAGE && UseShmem)
		PgShmemDestroy (memory);
	else
		free(memory);		
}